@@ -14,14 +14,18 @@
  * limitations under the License.
  */
 
-#include "system.h"
+#ifndef EXPEDITOR_EXTERNAL_USE
+# include "system.h"
+#endif
 
 #ifdef FEATURE_EXPEDITOR
 
 /* locally defined functions */
-static IBOOL s_ee_init_term(void);
+static IBOOL s_ee_init_term(iptr in, iptr out);
 static ptr s_ee_read_char(IBOOL blockp);
-static void s_ee_write_char(wchar_t c);
+static int s_ee_write_char(wchar_t c);
+static int s_ee_char_width(wchar_t c);
+static void s_ee_set_color(int color_id, IBOOL background);
 static void s_ee_flush(void);
 static ptr s_ee_get_screen_size(void);
 static void s_ee_raw(void);
@@ -53,14 +57,25 @@ static INT init_status = -1;
 static HANDLE hStdout, hStdin; 
 static DWORD InMode, OutMode;
 
-static IBOOL s_ee_init_term(void) {
-  CONSOLE_SCREEN_BUFFER_INFO csbiInfo; 
+static IBOOL s_ee_init_term(iptr hIn, iptr hOut) {
+  CONSOLE_SCREEN_BUFFER_INFO csbiInfo;
 
-  if (init_status != -1) return init_status;
+  if (hIn == -1)
+    hIn = (iptr)GetStdHandle(STD_INPUT_HANDLE);
+  if (hOut == -1)
+    hOut = (iptr)GetStdHandle(STD_OUTPUT_HANDLE);
+
+  if (init_status != -1) {
+    if ((HANDLE)hIn != hStdin)
+      return 0;
+    if ((HANDLE)hOut != hStdout)
+      return 0;
+    return init_status;
+  }
 
   init_status =
-     (hStdin = GetStdHandle(STD_INPUT_HANDLE)) != INVALID_HANDLE_VALUE
-     && (hStdout = GetStdHandle(STD_OUTPUT_HANDLE)) != INVALID_HANDLE_VALUE
+    (hStdin = (HANDLE)hIn) != INVALID_HANDLE_VALUE
+    && (hStdout = (HANDLE)hOut) != INVALID_HANDLE_VALUE
      && GetConsoleScreenBufferInfo(hStdout, &csbiInfo)
      && GetConsoleMode(hStdin, &InMode)
      && GetConsoleMode(hStdout, &OutMode);
@@ -77,7 +92,7 @@ static ptr s_ee_read_char(IBOOL blockp) {
   ptr tc;
 #endif /* PTHREADS */
   BOOL succ;
-  static char buf[10];
+  static wchar_t buf[20];
   static int bufidx = 0;
   static int buflen = 0;
   static int rptcnt = 0;
@@ -95,7 +110,7 @@ static ptr s_ee_read_char(IBOOL blockp) {
     if (!blockp) {
        DWORD NumberOfEvents;
        if (!GetNumberOfConsoleInputEvents(hStdin, &NumberOfEvents))
-         S_error1("expeditor", "error getting console info: ~a",
+         S_error1("expeditor", "error getting console input: ~a",
                     S_LastErrorString());
        if (NumberOfEvents == 0) return Sfalse;
     }
@@ -104,13 +119,13 @@ static ptr s_ee_read_char(IBOOL blockp) {
     tc = get_thread_context();
     if (DISABLECOUNT(tc) == FIX(0)) {
         deactivate_thread(tc);
-        succ = ReadConsoleInput(hStdin, irInBuf, 1, &cNumRead);
+        succ = ReadConsoleInputW(hStdin, irInBuf, 1, &cNumRead);
         reactivate_thread(tc);
     } else {
-        succ = ReadConsoleInput(hStdin, irInBuf, 1, &cNumRead);
+        succ = ReadConsoleInputW(hStdin, irInBuf, 1, &cNumRead);
     }
 #else /* PTHREADS */
-    succ = ReadConsoleInput(hStdin, irInBuf, 1, &cNumRead);
+    succ = ReadConsoleInputW(hStdin, irInBuf, 1, &cNumRead);
 #endif /* PTHREADS */
 
 
@@ -125,10 +140,10 @@ static ptr s_ee_read_char(IBOOL blockp) {
         KEY_EVENT_RECORD ker = irInBuf[0].Event.KeyEvent; 
         rptcnt = ker.wRepeatCount;
         if (ker.bKeyDown) {
-          char c;
+          wchar_t c;
 
-          if (c = ker.uChar.AsciiChar) {
-           /* translate ^@ 2) and ^<space> to nul */
+          if ((c = ker.uChar.UnicodeChar)) {
+            /* translate ^<space> to nul */
             if (c == 0x20 && (ker.dwControlKeyState & (LEFT_CTRL_PRESSED|RIGHT_CTRL_PRESSED)))
               buf[0] = 0;
             else
@@ -212,6 +227,91 @@ static ptr s_ee_read_char(IBOOL blockp) {
        when the window size changes. */
       case WINDOW_BUFFER_SIZE_EVENT: // scrn buf. resizing 
         return Strue;
+
+    case MOUSE_EVENT: {
+      MOUSE_EVENT_RECORD mer = irInBuf[0].Event.MouseEvent;
+      short x = mer.dwMousePosition.X % 1000;
+      short y = mer.dwMousePosition.Y % 1000;
+      short bstate = mer.dwButtonState & 0xffff;
+      short hbstate = mer.dwButtonState >> 16;
+      static short obstate = 0;
+
+      /* Mouse reporting starts with CSI [ < */
+      bufidx = 0;
+      buf[bufidx++] = '\033';
+      buf[bufidx++] = '[';
+      buf[bufidx++] = '<';
+
+      /* button mask (+ extra information) */
+      int bmask = 32;
+      char action = 'm';
+
+      if ((bstate & 1) ^ (obstate & 1)) {
+	/* left button changed, keep bmask 0, remove possible movement flag */
+	if (bstate & 1) {
+	  action = 'M';
+	}
+	bmask = 0;
+	obstate = (obstate & ~1) | (bstate & 1);
+      } else if ((bstate & 2) ^ (obstate & 2)) {
+	/* right button */
+	if (bstate & 2) {
+	  action = 'M';
+	}
+	bmask = 2;
+	obstate = (obstate & ~2) | (bstate & 2);
+      } else if ((bstate & 4) ^ (obstate & 4)) {
+	/* middle (second left) button */
+	if (bstate & 4) {
+	  action = 'M';
+	}
+	bmask = 1;
+	obstate = (obstate & ~4) | (bstate & 4);
+      } else if (mer.dwEventFlags == 4) {
+	/* MOUSE_WHEELED */
+	action = 'M';
+	if (hbstate > 0) {
+	  /* positive: wheel down */
+	  bmask = 65;
+	} else {
+	  /* negative: wheel up */
+	  bmask = 64;
+	}
+      }
+      if (bmask >= 10) {
+	buf[bufidx++] = '0' + (bmask / 10);
+      }
+      buf[bufidx++] = '0' + (bmask % 10);
+      buf[bufidx++] = ';';
+
+      /* X coordinate */
+      if (x >= 100) {
+	buf[bufidx++] = '0' + (x / 100);
+      }
+      if (x >= 10) {
+	buf[bufidx++] = '0' + ((x / 10) % 10);
+      }
+      buf[bufidx++] = '0' + (x % 10);
+      buf[bufidx++] = ';';
+
+      /* Y coordinate */
+      if (y >= 100) {
+	buf[bufidx++] = '0' + (y / 100);
+      }
+      if (y >= 10) {
+	buf[bufidx++] = '0' + ((y / 10) % 10);
+      }
+      buf[bufidx++] = '0' + (y % 10);
+
+      /* terminate with m/M based on button state changes */
+      buf[bufidx++] = action;
+
+      /* send however many characters were needed to encode the event */
+      buflen = bufidx;
+      bufidx = 0;
+      rptcnt = 1;
+      break;
+    }
   
       default: 
         break; 
@@ -246,6 +346,21 @@ static void s_ee_noraw(void) {
                S_LastErrorString());
 }
 
+static void s_ee_postoutput(void) { }
+static void s_ee_nopostoutput(void) { }
+
+static void s_ee_signal(void) {
+  if (!SetConsoleMode(hStdin, ENABLE_WINDOW_INPUT | ENABLE_PROCESSED_INPUT))
+    S_error1("expeditor", "error setting signal mode: ~a",
+             S_LastErrorString());
+}
+      
+static void s_ee_nosignal(void) {
+  if (!SetConsoleMode(hStdin, ENABLE_WINDOW_INPUT))
+    S_error1("expeditor", "error setting nosignal mode: ~a",
+             S_LastErrorString());
+}
+
 static void s_ee_enter_am_mode(void) { return; }
 
 static void s_ee_exit_am_mode(void) { return; }
@@ -508,12 +623,15 @@ static ptr s_ee_get_clipboard(void) {
   ptr x = S_G.null_string;
 
   if (OpenClipboard((HWND)0)) {
-    HANDLE h = GetClipboardData(CF_TEXT);
-     
-    if (h != (HANDLE *)0) {
-      char *s = (char *)GlobalLock(h);
-      if (s != (char *)0) x = Sstring(s);
-      GlobalUnlock(h);
+    HANDLE h = GetClipboardData(CF_UNICODETEXT);
+    if (h != NULL) {
+      wchar_t *w = (wchar_t*)GlobalLock(h);
+      if (w != NULL) {
+        char *s = Swide_to_utf8(w);
+        x = Sstring_utf8(s, -1);
+        free(s);
+        GlobalUnlock(h);
+      }
     }
     CloseClipboard();
   }
@@ -521,40 +639,139 @@ static ptr s_ee_get_clipboard(void) {
   return x;
 }
 
-static void s_ee_write_char(wchar_t c) {
-  if (c > 255) c = '?';
-  putchar(c);
+static int s_ee_write_char(wchar_t c) {
+  DWORD n;
+  CONSOLE_SCREEN_BUFFER_INFO pre_info, post_info;
+
+  GetConsoleScreenBufferInfo(hStdout, &pre_info);
+
+  WriteConsoleW(hStdout, &c, 1, &n, NULL);
+
+  GetConsoleScreenBufferInfo(hStdout, &post_info);
+
+  if (post_info.dwCursorPosition.X == post_info.dwSize.X-1) {
+    /* We don't know whether the cursor advanced as much as it would
+       earlier in the line, so return -128 to mean "unknown". */
+    return -128;
+  }
+
+  return post_info.dwCursorPosition.X - pre_info.dwCursorPosition.X;
+}
+
+static int s_ee_char_width(UNUSED wchar_t c)
+{
+  /* There's no `wcwidth` on Windows. An editor can gather results
+     from `s_ee_write_char`, though. */
+  return 1;
+}
+
+static int foreground_colors[] =
+  { 0, /* Black = 0 */
+    FOREGROUND_RED, /* Red = 1 */
+    FOREGROUND_GREEN, /* Green = ... */
+    FOREGROUND_RED | FOREGROUND_GREEN, /* Yellow */
+    FOREGROUND_BLUE, /* Blue */
+    FOREGROUND_RED | FOREGROUND_BLUE, /* Magenta */
+    FOREGROUND_GREEN | FOREGROUND_BLUE, /* Cyan */
+    FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE, /* Light gray */
+    FOREGROUND_INTENSITY, /* Dark gray */
+    FOREGROUND_RED | FOREGROUND_INTENSITY, /* Light red */
+    FOREGROUND_GREEN | FOREGROUND_INTENSITY, /* Light green*/
+    FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY, /* Yellow */
+    FOREGROUND_BLUE | FOREGROUND_INTENSITY, /* Blue */
+    FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY, /* Magenta */
+    FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY, /* Cyan */
+    FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY, /* Light gray */
+  };
+
+static int background_colors[] =
+  { 0, /* Black = 0 */
+    BACKGROUND_RED, /* Red = 1 */
+    BACKGROUND_GREEN, /* Green = ... */
+    BACKGROUND_RED | BACKGROUND_GREEN, /* Yellow */
+    BACKGROUND_BLUE, /* Blue */
+    BACKGROUND_RED | BACKGROUND_BLUE, /* Magenta */
+    BACKGROUND_GREEN | BACKGROUND_BLUE, /* Cyan */
+    BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE, /* Light gray */
+    BACKGROUND_INTENSITY, /* Dark gray */
+    BACKGROUND_RED | BACKGROUND_INTENSITY, /* Light red */
+    BACKGROUND_GREEN | BACKGROUND_INTENSITY, /* Light green*/
+    BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_INTENSITY, /* Yellow */
+    BACKGROUND_BLUE | BACKGROUND_INTENSITY, /* Blue */
+    BACKGROUND_RED | BACKGROUND_BLUE | BACKGROUND_INTENSITY, /* Magenta */
+    BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY, /* Cyan */
+    BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE | BACKGROUND_INTENSITY, /* Light gray */
+  };
+
+static int initial_colors_set = 0;
+static int current_foreground, initial_foreground;
+static int current_background, initial_background;
+
+static void s_ee_set_color(int color_id, IBOOL background) {
+  if (!initial_colors_set) {
+    CONSOLE_SCREEN_BUFFER_INFO info;
+    int i;
+    GetConsoleScreenBufferInfo(hStdout, &info);
+    for (i = 0; i < 16; i++) {
+      if ((info.wAttributes & 0xf) == foreground_colors[i])
+        initial_foreground = i;
+      if ((info.wAttributes & 0xf0) == background_colors[i])
+        initial_background = i;
+    }
+    initial_colors_set = 1;
+    current_foreground = initial_foreground;
+    current_background = initial_background;
+  }
+
+  if (color_id < 0) {
+    current_foreground = initial_foreground;
+    current_background = initial_background;
+  } else if (background)
+    current_background = color_id;
+  else
+    current_foreground = color_id;
+
+  SetConsoleTextAttribute(hStdout, (foreground_colors[current_foreground]
+                                    | background_colors[current_background]));
 }
 
 #else /* WIN32 */
+
 #include <limits.h>
 #ifdef DISABLE_CURSES
-#include "nocurses.h"
+# include "nocurses.h"
 #elif defined(SOLARIS)
-#define NCURSES_CONST
-#define CHTYPE int
-#include </usr/include/curses.h>
-#include </usr/include/term.h>
-#elif defined(NETBSD)
-#include <ncurses.h>
-#include <ncurses/term.h>
-#else /* NETBSD */
-#include <curses.h>
-#include <term.h>
-#endif /* SOLARIS */
+# define NCURSES_CONST
+# define CHTYPE int
+# include </usr/include/curses.h>
+# include </usr/include/term.h>
+#elif defined(__DragonFly__)
+# include <ncurses/curses.h>
+# include <ncurses/term.h>
+#else
+# include <curses.h>
+# include <term.h>
+#endif
 #include <termios.h>
 #include <signal.h>
-#include <time.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #include <wchar.h>
 #include <locale.h>
-#if !defined(__GLIBC__) && !defined(__OpenBSD__) && !defined(__NetBSD__)
-#include <xlocale.h>
+#if defined(__linux__)
+# include <unistd.h>
+# include <time.h>
+#endif
+#if !defined(__GLIBC__) && !defined(__COSMOPOLITAN__) && !defined(__OpenBSD__) && !defined(__NetBSD__) && !defined(__linux__) && !defined(__EMSCRIPTEN__) && !defined(NO_USELOCALE)
+# include <xlocale.h>
+#endif
+
+#if (defined(__gnu_hurd__) || defined(__linux__)) && !defined(_XOPEN_SOURCE)
+extern int wcwidth(wchar_t);
 #endif
 
 #if defined(TIOCGWINSZ) && defined(SIGWINCH) && defined(EINTR)
-#define HANDLE_SIGWINCH
+# define HANDLE_SIGWINCH
 #endif
 
 #ifdef USE_MBRTOWC_L
@@ -585,18 +802,41 @@ static void handle_sigwinch(UNUSED int sig) {
 #define STDOUT_FD 1
 
 static IBOOL disable_auto_margin = 0, avoid_last_column = 0;
+#ifndef NO_USELOCALE
 static locale_t term_locale;
+#endif
 static mbstate_t term_in_mbs;
 static mbstate_t term_out_mbs;
 
-static IBOOL s_ee_init_term(void) {
+static IBOOL s_ee_init_term(iptr in, iptr out) {
   int errret;
+#if defined(SCHEME_PORTABLE_TERM) && !defined(DISABLE_CURSES)
+  const char *term_env;
+#else
+# define term_env NULL
+#endif
 
+  if (in == -1) in = STDIN_FD;
+  if (out == -1) out = STDOUT_FD;
+
+  /* uses C library with stdout and stderr */
+  if (in != STDIN_FD) return 0;
+  if (out != STDOUT_FD) return 0;
+  
   if (init_status != -1) return init_status;
 
+#if defined(SCHEME_PORTABLE_TERM) && !defined(DISABLE_CURSES)
+  /* avoid problems with statically linked ncurses running on newer */
+  term_env = getenv("TERM");
+  if (term_env && !strcmp(term_env, "xterm-256color"))
+    term_env = "xterm";
+  else
+    term_env = NULL;
+#endif
+
   if (isatty(STDIN_FD)
       && isatty(STDOUT_FD)
-      && setupterm(NULL, STDOUT_FD, &errret) != ERR
+      && setupterm((char *)term_env, STDOUT_FD, &errret) != ERR
 /* assuming here and in our optproc definitions later that the names of
    missing capabilities are set to NULL, although this does not appear
    to be documented */
@@ -640,7 +880,9 @@ static IBOOL s_ee_init_term(void) {
     sigaction(SIGWINCH, &act, (struct sigaction *)0);
 #endif
 
+#ifndef NO_USELOCALE
     term_locale = newlocale(LC_ALL_MASK, "", NULL);
+#endif
     memset(&term_out_mbs, 0, sizeof(term_out_mbs));
     memset(&term_in_mbs, 0, sizeof(term_in_mbs));
 
@@ -656,7 +898,6 @@ static IBOOL s_ee_init_term(void) {
    only if blockp is false */
 static ptr s_ee_read_char(IBOOL blockp) {
   ptr msg; int fd = STDIN_FD; int n; char buf[1]; wchar_t wch; size_t sz;
-  locale_t old_locale;
 #ifdef PTHREADS
   ptr tc = get_thread_context();
 #endif
@@ -694,14 +935,19 @@ static ptr s_ee_read_char(IBOOL blockp) {
       if (buf[0] == '\0') {
         return Schar('\0');
       } else {
-        old_locale = uselocale(term_locale);
+#ifndef NO_USELOCALE
+        locale_t old_locale = uselocale(term_locale);
+#endif
         sz = mbrtowc(&wch, buf, 1, &term_out_mbs);
+#ifndef NO_USELOCALE
         uselocale(old_locale);
+#endif
         if (sz == 1) {
           return Schar(wch);
         }
       }
-    }
+    } else
+      sz = 0;
 
   } while ((n < 0 && errno == EINTR) || (n == 1 && sz == (size_t)-2));
 
@@ -729,7 +975,7 @@ static ptr s_ee_get_screen_size(void) {
   static IBOOL tried_resize = 0;
  /* attempt to work around 10.6 tty driver / xterm bug */
   if (ee_rows == 0 && ee_cols == 0 && !tried_resize) {
-    system("exec /usr/X11/bin/resize >& /dev/null");
+    SYSTEM("exec /usr/X11/bin/resize >& /dev/null");
     tried_resize = 1;
     return s_ee_get_screen_size();
   }
@@ -763,17 +1009,36 @@ static int eeputc(tputsputcchar c) {
 
 static struct termios orig_termios;
 
+static void attr_error (const char *who) {
+  char buf[256];
+  ptr msg = S_strerror(errno);
+  if (msg != Sfalse) {
+    snprintf(buf, sizeof(buf), "error entering %s mode: ~a", who);
+    S_error1("expeditor", buf, msg);
+  } else {
+    snprintf(buf, sizeof(buf), "error entering %s mode: ~a", who);
+    S_error("expeditor", buf);
+  }
+}
+
+static void get_stdin_attr (struct termios *t, const char *who) {
+  while (tcgetattr(STDIN_FD, t) != 0) {
+    if (errno != EINTR)
+      attr_error(who);
+  }
+}
+
+static void set_stdin_attr (struct termios *t, const char *who) {
+  while (tcsetattr(STDIN_FD, TCSADRAIN, t) != 0) {
+    if (errno != EINTR)
+      attr_error(who);
+  }
+}
+
 static void s_ee_raw(void) {
   struct termios new_termios;
-  while (tcgetattr(STDIN_FD, &orig_termios) != 0) {
-    if (errno != EINTR) {
-      ptr msg = S_strerror(errno);
-      if (msg != Sfalse)
-        S_error1("expeditor", "error entering raw mode: ~a", msg);
-      else
-        S_error("expeditor", "error entering raw mode");
-    }
-  }
+  get_stdin_attr(&orig_termios, "raw");
+  
   new_termios = orig_termios;
 
  /* essentially want "stty raw -echo".  the appropriate flags to accomplish
@@ -788,27 +1053,47 @@ static void s_ee_raw(void) {
   new_termios.c_cc[VMIN] = 1;
   new_termios.c_cc[VTIME] = 0;
 
-  while (tcsetattr(STDIN_FD, TCSADRAIN, &new_termios) != 0) {
-    if (errno != EINTR) {
-      ptr msg = S_strerror(errno);
-      if (msg != Sfalse)
-        S_error1("expeditor", "error entering raw mode: ~a", msg);
-      else
-        S_error("expeditor", "error entering raw mode");
-    }
-  }
+  set_stdin_attr(&new_termios, "raw");
 }
 
 static void s_ee_noraw(void) {
-  while (tcsetattr(STDIN_FD, TCSADRAIN, &orig_termios) != 0) {
-    if (errno != EINTR) {
-      ptr msg = S_strerror(errno);
-      if (msg != Sfalse)
-        S_error1("expeditor", "error leaving raw mode: ~a", msg);
-      else
-        S_error("expeditor", "error leaving raw mode");
-    }
-  }
+  set_stdin_attr(&orig_termios, "noraw");
+}
+
+static void s_ee_postoutput(void) {
+  struct termios new_termios;
+  get_stdin_attr(&new_termios, "postoutput");
+
+  new_termios.c_oflag |= OPOST;
+
+  set_stdin_attr(&new_termios, "postoutput");
+}
+
+static void s_ee_nopostoutput(void) {
+  struct termios new_termios;
+  get_stdin_attr(&new_termios, "nopostoutput");
+
+  new_termios.c_oflag &= (~OPOST);
+
+  set_stdin_attr(&new_termios, "nopostoutput");
+}
+
+static void s_ee_signal(void) {
+  struct termios new_termios;
+  get_stdin_attr(&new_termios, "signal");
+
+  new_termios.c_lflag |= ISIG;
+
+  set_stdin_attr(&new_termios, "signal");
+}
+
+static void s_ee_nosignal(void) {
+  struct termios new_termios;
+  get_stdin_attr(&new_termios, "nosignal");
+
+  new_termios.c_lflag &= (~ISIG);
+
+  set_stdin_attr(&new_termios, "nosignal");
 }
 
 static void s_ee_enter_am_mode(void) {
@@ -1037,17 +1322,65 @@ static ptr s_ee_get_clipboard(void) {
   return p;
 }
 
-static void s_ee_write_char(wchar_t wch) {
-  locale_t old; char buf[MB_LEN_MAX]; size_t n;
+static int s_ee_write_char(wchar_t wch) {
+  int width;
+  char buf[MB_LEN_MAX]; size_t n;
+#ifndef NO_USELOCALE
+  locale_t old = uselocale(term_locale);
+#endif
 
-  old = uselocale(term_locale);
   n = wcrtomb(buf, wch, &term_in_mbs);
   if (n == (size_t)-1) {
     putchar('?');
   } else {
     fwrite(buf, 1, n, stdout);
   }
+
+  width = wcwidth(wch);
+
+#ifndef NO_USELOCALE
   uselocale(old);
+#endif
+
+  return width;
+}
+
+static int s_ee_char_width(wchar_t wch)
+{
+  int width;
+#ifndef NO_USELOCALE
+  locale_t old = uselocale(term_locale);
+#endif
+
+  width = wcwidth(wch);
+
+#ifndef NO_USELOCALE
+  uselocale(old);
+#endif
+
+  return width;
+}
+
+/* see Windows s_ee_set_color for color-index meanings */
+static void s_ee_set_color(int color_id, IBOOL background) {
+  char buf[6];
+  int len = 5;
+  memcpy(buf, "\033[__m", len);
+  if (color_id < 0) {
+    buf[2] = '0';
+    buf[3] = 'm';
+    len = 4;
+  } else if (background && (color_id > 8)) {
+    buf[2] = '1';
+    buf[3] = '0';
+    buf[4] = '0' + (color_id & 0x7);
+    buf[5] = 'm';
+    len = 6;
+  } else {
+    buf[2] = ((color_id > 8) ? '9' : (background ? '4' : '3'));
+    buf[3] = '0' + (color_id & 0x7);
+  }
+  fwrite(buf, 1, len, stdout);
 }
 
 #endif /* WIN32 */
@@ -1056,16 +1389,32 @@ static void s_ee_flush(void) {
   fflush(stdout);
 }
 
+static ptr s_ee_pending_winch() {
+#ifdef HANDLE_SIGWINCH
+  return winched ? Strue : Sfalse;
+#else
+  return Sfalse;
+#endif
+}
+
 void S_expeditor_init(void) {
   Sforeign_symbol("(cs)ee_init_term", (void *)s_ee_init_term);
   Sforeign_symbol("(cs)ee_read_char", (void *)s_ee_read_char);
+  Sforeign_symbol("(cs)ee_pending_winch", (void *)s_ee_pending_winch);
   Sforeign_symbol("(cs)ee_write_char", (void *)s_ee_write_char);
+  Sforeign_symbol("(cs)ee_char_width", (void *)s_ee_char_width);
+  Sforeign_symbol("(cs)ee_set_color", (void *)s_ee_set_color);
   Sforeign_symbol("(cs)ee_flush", (void *)s_ee_flush);
   Sforeign_symbol("(cs)ee_get_screen_size", (void *)s_ee_get_screen_size);
   Sforeign_symbol("(cs)ee_raw", (void *)s_ee_raw);
   Sforeign_symbol("(cs)ee_noraw", (void *)s_ee_noraw);
+  Sforeign_symbol("(cs)ee_postoutput", (void *)s_ee_postoutput);
+  Sforeign_symbol("(cs)ee_nopostoutput", (void *)s_ee_nopostoutput);
+  Sforeign_symbol("(cs)ee_signal", (void *)s_ee_signal);
+  Sforeign_symbol("(cs)ee_nosignal", (void *)s_ee_nosignal);
   Sforeign_symbol("(cs)ee_enter_am_mode", (void *)s_ee_enter_am_mode);
   Sforeign_symbol("(cs)ee_exit_am_mode", (void *)s_ee_exit_am_mode);
+  Sforeign_symbol("(cs)ee_set_color", (void *)s_ee_set_color);
   Sforeign_symbol("(cs)ee_pause", (void *)s_ee_pause);
   Sforeign_symbol("(cs)ee_nanosleep", (void *)s_ee_nanosleep);
   Sforeign_symbol("(cs)ee_get_clipboard", (void *)s_ee_get_clipboard);
